#!/usr/bin/env python

# Copyright 2015 The Kubernetes Authors All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function

import json
import mmap
import os
import re
import sys

def PrintError(*err):
  print(*err, file=sys.stderr)

def file_passes(filename, extension, ref, regexs):
    try:
        f = open(filename, 'r')
    except:
        return False

    data = f.read()

    # remove build tags from the top of Go file
    if extension == "go":
        p = regexs["go_build_constraints"]
        (data, found) = p.subn("", data, 1)

    data = data.splitlines()

    # if our test file is smaller than the reference it surely fails!
    if len(ref) > len(data):
        return False

    # trim our file to the same number of lines as the reference file
    data = data[:len(ref)]

    p = regexs["year"]
    for d in data:
        if p.search(d):
            return False

    # Replace all occurances of the regex "2015|2014" with "YEAR"
    p = regexs["date"]
    for i, d in enumerate(data):
        (data[i], found) = p.subn('YEAR', d)
        if found != 0:
            break

    # if we don't match the reference at this point, fail
    if ref != data:
        return False

    return True

def main():
    if len(sys.argv) < 3:
        PrintError("usage: %s extension FILENAME [FILENAMES]" % sys.argv[0])
        return False

    basedir = os.path.dirname(os.path.abspath(__file__))

    extension = sys.argv[1]
    # argv[0] is the binary, argv[1] is the extension (go, sh, py, whatever)
    filenames = sys.argv[2:]

    ref_filename = basedir + "/boilerplate." + extension + ".txt"
    try:
        ref_file = open(ref_filename, 'r')
    except:
        # No boilerplate template is success
        return True
    ref = ref_file.read().splitlines()

    regexs = {}
    # Search for "YEAR" which exists in the boilerplate, but shouldn't in the real thing
    regexs["year"] = re.compile( 'YEAR' )
    # dates can be 2014 or 2015, company holder names can be anything
    regexs["date"] = re.compile( '(2014|2015)' )
    # strip // +build \n\n build constraints
    regexs["go_build_constraints"] = re.compile(r"^(// \+build.*\n)+\n", re.MULTILINE)

    for filename in filenames:
        if not file_passes(filename, extension, ref, regexs):
            print(filename, file=sys.stdout)

if __name__ == "__main__":
  sys.exit(main())



#!/usr/bin/env python

# Copyright 2015 The Kubernetes Authors All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
from mock import MagicMock, patch, call
from path import Path
import pytest
import sys

d = Path('__file__').parent.abspath() / 'hooks'
sys.path.insert(0, d.abspath())

from lib.registrator import Registrator

class TestRegistrator():

    def setup_method(self, method):
        self.r = Registrator()

    def test_data_type(self):
        if type(self.r.data) is not dict:
            pytest.fail("Invalid type")

    @patch('json.loads')
    @patch('httplib.HTTPConnection')
    def test_register(self, httplibmock, jsonmock):
        result = self.r.register('foo', 80, '/v1/test')

        httplibmock.assert_called_with('foo', 80)
        requestmock = httplibmock().request
        requestmock.assert_called_with(
                "POST", "/v1/test",
                json.dumps(self.r.data),
                {"Content-type": "application/json",
                 "Accept": "application/json"})


    def test_command_succeeded(self):
        response = MagicMock()
        result = json.loads('{"status": "Failure", "kind": "Status", "code": 409, "apiVersion": "v1", "reason": "AlreadyExists", "details": {"kind": "node", "name": "10.200.147.200"}, "message": "node \\"10.200.147.200\\" already exists", "creationTimestamp": null}')
        response.status = 200
        self.r.command_succeeded(response, result)
        response.status = 500
        with pytest.raises(RuntimeError):
            self.r.command_succeeded(response, result)
        response.status = 409
        with pytest.raises(ValueError):
            self.r.command_succeeded(response, result)



#!/usr/bin/env python

# Copyright 2015 The Kubernetes Authors All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
The main hook file is called by Juju.
"""
import contextlib
import os
import socket
import subprocess
import sys
from charmhelpers.core import hookenv, host
from kubernetes_installer import KubernetesInstaller
from path import path

hooks = hookenv.Hooks()


@contextlib.contextmanager
def check_sentinel(filepath):
    """
    A context manager method to write a file while the code block is doing
    something and remove the file when done.
    """
    fail = False
    try:
        yield filepath.exists()
    except:
        fail = True
        filepath.touch()
        raise
    finally:
        if fail is False and filepath.exists():
            filepath.remove()


@hooks.hook('config-changed')
def config_changed():
    """
    On the execution of the juju event 'config-changed' this function
    determines the appropriate architecture and the configured version to
    create kubernetes binary files.
    """
    hookenv.log('Starting config-changed')
    charm_dir = path(hookenv.charm_dir())
    config = hookenv.config()
    # Get the version of kubernetes to install.
    version = config['version']

    if version == 'master':
        # The 'master' branch of kuberentes is used when master is configured.
        branch = 'master'
    elif version == 'local':
        # Check for kubernetes binaries in the local files/output directory.
        branch = None
    else:
        # Create a branch to a tag to get the release version.
        branch = 'tags/{0}'.format(version)

    # Get the package architecture, rather than arch from the kernel (uname -m).
    arch = subprocess.check_output(['dpkg', '--print-architecture']).strip()

    if not branch:
        output_path = charm_dir / 'files/output'
        installer = KubernetesInstaller(arch, version, output_path)
    else:

        # Build the kuberentes binaries from source on the units.
        kubernetes_dir = path('/opt/kubernetes')

        # Construct the path to the binaries using the arch.
        output_path = kubernetes_dir / '_output/local/bin/linux' / arch
        installer = KubernetesInstaller(arch, version, output_path)

        if not kubernetes_dir.exists():
            print('The source directory {0} does not exist'.format(kubernetes_dir))
            print('Was the kubernetes code cloned during install?')
            exit(1)

        # Change to the kubernetes directory (git repository).
        with kubernetes_dir:

            # Create a command to get the current branch.
            git_branch = 'git branch | grep "\*" | cut -d" " -f2'
            current_branch = subprocess.check_output(git_branch, shell=True).strip()
            print('Current branch: ', current_branch)
            # Create the path to a file to indicate if the build was broken.
            broken_build = charm_dir / '.broken_build'
            # write out the .broken_build file while this block is executing.
            with check_sentinel(broken_build) as last_build_failed:
                print('Last build failed: ', last_build_failed)
                # Rebuild if current version is different or last build failed.
                if current_branch != version or last_build_failed:
                    installer.build(branch)
            if not output_path.isdir():
                broken_build.touch()

    # Create the symoblic links to the right directories.
    installer.install()

    relation_changed()

    hookenv.log('The config-changed hook completed successfully.')


@hooks.hook('etcd-relation-changed', 'minions-api-relation-changed')
def relation_changed():
    template_data = get_template_data()

    # Check required keys
    for k in ('etcd_servers',):
        if not template_data.get(k):
            print "Missing data for", k, template_data
            return

    print "Running with\n", template_data

    # Render and restart as needed
    for n in ('apiserver', 'controller-manager', 'scheduler'):
        if render_file(n, template_data) or not host.service_running(n):
            host.service_restart(n)

    # Render the file that makes the kubernetes binaries available to minions.
    if render_file(
            'distribution', template_data,
            'conf.tmpl', '/etc/nginx/sites-enabled/distribution') or \
            not host.service_running('nginx'):
        host.service_reload('nginx')
    # Render the default nginx template.
    if render_file(
            'nginx', template_data,
            'conf.tmpl', '/etc/nginx/sites-enabled/default') or \
            not host.service_running('nginx'):
        host.service_reload('nginx')

    # Send api endpoint to minions
    notify_minions()


@hooks.hook('network-relation-changed')
def network_relation_changed():
    relation_id = hookenv.relation_id()
    hookenv.relation_set(relation_id, ignore_errors=True)


def notify_minions():
    print("Notify minions.")
    config = hookenv.config()
    for r in hookenv.relation_ids('minions-api'):
        hookenv.relation_set(
            r,
            hostname=hookenv.unit_private_ip(),
            port=8080,
            version=config['version'])
    print("Notified minions of version " + config['version'])


def get_template_data():
    rels = hookenv.relations()
    config = hookenv.config()
    version = config['version']
    template_data = {}
    template_data['etcd_servers'] = ",".join([
        "http://%s:%s" % (s[0], s[1]) for s in sorted(
            get_rel_hosts('etcd', rels, ('hostname', 'port')))])
    template_data['minions'] = ",".join(get_rel_hosts('minions-api', rels))

    template_data['api_bind_address'] = _bind_addr(hookenv.unit_private_ip())
    template_data['bind_address'] = "127.0.0.1"
    template_data['api_server_address'] = "http://%s:%s" % (
        hookenv.unit_private_ip(), 8080)
    arch = subprocess.check_output(['dpkg', '--print-architecture']).strip()

    template_data['web_uri'] = "/kubernetes/%s/local/bin/linux/%s/" % (version,
                                                                       arch)
    if version == 'local':
        template_data['alias'] = hookenv.charm_dir() + '/files/output/'
    else:
        directory = '/opt/kubernetes/_output/local/bin/linux/%s/' % arch
        template_data['alias'] = directory
    _encode(template_data)
    return template_data


def _bind_addr(addr):
    if addr.replace('.', '').isdigit():
        return addr
    try:
        return socket.gethostbyname(addr)
    except socket.error:
            raise ValueError("Could not resolve private address")


def _encode(d):
    for k, v in d.items():
        if isinstance(v, unicode):
            d[k] = v.encode('utf8')


def get_rel_hosts(rel_name, rels, keys=('private-address',)):
    hosts = []
    for r, data in rels.get(rel_name, {}).items():
        for unit_id, unit_data in data.items():
            if unit_id == hookenv.local_unit():
                continue
            values = [unit_data.get(k) for k in keys]
            if not all(values):
                continue
            hosts.append(len(values) == 1 and values[0] or values)
    return hosts


def render_file(name, data, src_suffix="upstart.tmpl", tgt_path=None):
    tmpl_path = os.path.join(
        os.environ.get('CHARM_DIR'), 'files', '%s.%s' % (name, src_suffix))

    with open(tmpl_path) as fh:
        tmpl = fh.read()
    rendered = tmpl % data

    if tgt_path is None:
        tgt_path = '/etc/init/%s.conf' % name

    if os.path.exists(tgt_path):
        with open(tgt_path) as fh:
            contents = fh.read()
        if contents == rendered:
            return False

    with open(tgt_path, 'w') as fh:
        fh.write(rendered)
    return True

if __name__ == '__main__':
    hooks.execute(sys.argv)



#!/usr/bin/env python

# Copyright 2014 The Kubernetes Authors All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# This is a very simple utility that reads a JSON document from stdin, parses it
# and returns the specified value.  The value is described using a simple dot
# notation.  If any errors are encountered along the way, an error is output and
# a failure value is returned.

from __future__ import print_function

import json
import sys

def PrintError(*err):
  print(*err, file=sys.stderr)

def main():
  try:
    obj = json.load(sys.stdin)
  except Exception, e:
    PrintError("Error loading JSON: {0}".format(str(e)))

  if len(sys.argv) == 1:
    # if we don't have a query string, return success
    return 0
  elif len(sys.argv) > 2:
    PrintError("Usage: {0} <json query>".format(sys.args[0]))
    return 1

  query_list = sys.argv[1].split('.')
  for q in query_list:
    if isinstance(obj, dict):
      if q not in obj:
        PrintError("Couldn't find '{0}' in dict".format(q))
        return 1
      obj = obj[q]
    elif isinstance(obj, list):
      try:
        index = int(q)
      except:
        PrintError("Can't use '{0}' to index into array".format(q))
        return 1
      if index >= len(obj):
        PrintError("Index ({0}) is greater than length of list ({1})".format(q, len(obj)))
        return 1
      obj = obj[index]
    else:
      PrintError("Trying to query non-queryable object: {0}".format(q))
      return 1

  if isinstance(obj, str):
    print(obj)
  else:
    print(json.dumps(obj, indent=2))

if __name__ == "__main__":
  sys.exit(main())



#!/usr/bin/env python

# Copyright 2015 The Kubernetes Authors All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Script to print out PR info in release note format.

import json
import sys
import urllib2

PULLQUERY=("https://api.github.com/repos/"
           "GoogleCloudPlatform/kubernetes/pulls/{pull}")
LOGIN="login"
TITLE="title"
USER="user"

def print_pulls(pulls):
  for pull in pulls:
    d = json.loads(urllib2.urlopen(PULLQUERY.format(pull=pull)).read())
    print "* {title} #{pull} ({author})".format(
        title=d[TITLE], pull=pull, author=d[USER][LOGIN])

if __name__ == "__main__":
  if len(sys.argv) < 2:
    print ("Usage: {cmd} <pulls>...: Prints out short " +
           "markdown description for PRs appropriate for release notes.")
    sys.exit(1)
  print_pulls(sys.argv[1:])



#!/usr/bin/env python

# Copyright 2015 The Kubernetes Authors All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from mock import patch
from path import path
from path import Path
import pytest
import subprocess
import sys

# Add the hooks directory to the python path.
hooks_dir = Path('__file__').parent.abspath() / 'hooks'
sys.path.insert(0, hooks_dir.abspath())
# Import the module to be tested.
import kubernetes_installer


def test_run():
    """ Test the run method both with valid commands and invalid commands. """
    ls = 'ls -l {0}/kubernetes_installer.py'.format(hooks_dir)
    output = kubernetes_installer.run(ls, False)
    assert output
    assert 'kubernetes_installer.py' in output
    output = kubernetes_installer.run(ls, True)
    assert output
    assert 'kubernetes_installer.py' in output

    invalid_directory = path('/not/a/real/directory')
    assert not invalid_directory.exists()
    invalid_command = 'ls {0}'.format(invalid_directory)
    with pytest.raises(subprocess.CalledProcessError) as error:
        kubernetes_installer.run(invalid_command)
        print(error)
    with pytest.raises(subprocess.CalledProcessError) as error:
        kubernetes_installer.run(invalid_command, shell=True)
        print(error)


class TestKubernetesInstaller():

    def makeone(self, *args, **kw):
        """ Create the KubernetesInstaller object and return it. """
        from kubernetes_installer import KubernetesInstaller
        return KubernetesInstaller(*args, **kw)

    def test_init(self):
        """ Test that the init method correctly assigns the variables. """
        ki = self.makeone('i386', '3.0.1', '/tmp/does_not_exist')
        assert ki.aliases
        assert 'kube-apiserver' in ki.aliases
        assert 'kube-controller-manager' in ki.aliases
        assert 'kube-scheduler' in ki.aliases
        assert 'kubectl' in ki.aliases
        assert 'kubelet' in ki.aliases
        assert ki.arch == 'i386'
        assert ki.version == '3.0.1'
        assert ki.output_dir == path('/tmp/does_not_exist')

    @patch('kubernetes_installer.run')
    @patch('kubernetes_installer.subprocess.call')
    def test_build(self, cmock, rmock):
        """ Test the build method with master and non-master branches. """
        directory = path('/tmp/kubernetes_installer_test/build')
        ki = self.makeone('amd64', 'v99.00.11', directory)
        assert not directory.exists(), 'The %s directory exists!' % directory
        # Call the build method with "master" branch.
        ki.build("master")
        # TODO: run is called many times but mock only remembers last one.
        rmock.assert_called_with('git reset --hard origin/master')
        # TODO: call is complex and hard to verify with mock, fix that.
        cmock.assert_called_once()

        # Call the build method with something other than "master" branch.
        ki.build("branch")
        # TODO: run is called many times, but mock only remembers last one.
        rmock.assert_called_with('git checkout -b v99.00.11 branch')
        # TODO: call is complex and hard to verify with mock, fix that.
        cmock.assert_called_once()

        directory.rmtree_p()

    def test_install(self):
        """ Test the install method that it creates the correct links. """
        directory = path('/tmp/kubernetes_installer_test/install')
        ki = self.makeone('ppc64le', '1.2.3', directory)
        assert not directory.exists(), 'The %s directory exits!' % directory
        directory.makedirs_p()
        # Create the files for the install method to link to.
        (directory / 'kube-apiserver').touch()
        (directory / 'kube-controller-manager').touch()
        (directory / 'kube-proxy').touch()
        (directory / 'kube-scheduler').touch()
        (directory / 'kubectl').touch()
        (directory / 'kubelet').touch()

        results = directory / 'install/results/go/here'
        assert not results.exists()
        ki.install(results)
        assert results.isdir()
        # Check that all the files were correctly aliased and are links.
        assert (results / 'apiserver').islink()
        assert (results / 'controller-manager').islink()
        assert (results / 'kube-proxy').islink()
        assert (results / 'scheduler').islink()
        assert (results / 'kubectl').islink()
        assert (results / 'kubelet').islink()

        directory.rmtree_p()
